package webcore.utilities;

/*****
 * FileUtils class contains several static utility functions that are helpful
 * for use with files.
 */

public class FileUtils
{


	/***
	 * Determines if a file exists
	 * @return
	 * true if file exists<br>
	 * false if file does not exist
	 */
	public static boolean doesFileExist(String filePath)
 	{
  		java.io.File f = new java.io.File(filePath);
    	return f.exists();
  	}


	/****
	 * Reads the entire contents of a file and returns the contents as a String<p>
	 * This function should not be used on large files, because it reads the
	 * entire contents into memory as a String. It is meant to be used on very
	 * small text files.<p>
 	 * Its counterpart is writeFileAsString();
	 * @return
	 * the entire contents of the file as a String
	 */
 	public static String readFileAsString(String filePath)
  		throws java.io.IOException
  	{
   		return readFileAsString(new java.io.File(filePath));
   	}

	/****
	 * Reads the entire contents of a file and returns the contents as a String<p>
	 * This function should not be used on large files, because it reads the
	 * entire contents into memory as a String. It is meant to be used on very
	 * small text files.<p>
 	 * Its counterpart is writeFileAsString();
	 * @return
	 * the entire contents of the file as a String
	 */
 	public static String readFileAsString(java.io.File file) throws java.io.IOException
  	{
  		java.io.DataInputStream in = new java.io.DataInputStream(
    		new java.io.FileInputStream(file));
    	byte b[] = new byte[in.available()];
     	in.readFully(b);
    	return new String(b);
   	}

	/****
	 * Writes the String filedata to a file. This is a convenience function to
	 * to write a a short text file.<p>
 	 * Its counterpart is readFileAsString();
	 */
 	public static void writeFileAsString(String filePath, String filedata)
  		throws java.io.IOException
  	{
   		writeFileAsString(new java.io.File(filePath), filedata);
   	}

	/****
	 * Writes the String filedata to a file. This is a convenience function to
	 * to write a a short text file.<p>
 	 * Its counterpart is readFileAsString();
	 */
 	public static void writeFileAsString(java.io.File file, String filedata)
  		throws java.io.IOException
  	{
  		java.io.Writer out = new java.io.FileWriter(file);
    	out.write(filedata);
     	out.close();
   	}


	/****
	 * Extracts the key value of a property String of the form: <KEY>=<VALUE><p>
  	 * This method is used internally to read property files.<p>
	 * If the String contains no "=" the entire String is returned.<br>
	 * If the String starts with "=" a zero length String is returned.
	 * @return
	 * the key of a property String
	 */
    public static String extractKey(String propertyString)
    {
    	int x = propertyString.indexOf("=");
     	if (x == -1) return propertyString;
      	else if (x == 0) return "";
      	else return propertyString.substring(0, x-1);
    }


	/****
	 * Extracts the value value of a property String of the form: <KEY>=<VALUE><p>
  	 * This method is used internally to read property files.<p>
	 * If the String contains no "=" a zero length String is returned.<br>
	 * If the String starts with "=" the entire String is after the "=" is returned.
	 * @return
	 * the value of a property String
	 */
    public static String extractValue(String propertyString)
    {
    	int x = propertyString.indexOf("=");
     	if (x == -1) return "";
      	else if (x == 0) return propertyString;
      	else return propertyString.substring(x+1);
    }

    /*****
     * Loads the contents of a properties file which may contain
     * import statements for other properties files and returns a new
     * java.util.Properties object.  The properties file must be of the form
     * created by java.util.Properties.save(). The file is read using the same
     * rules as java.util.Properties.load() method with the addition of
     * <I>File Input Statements</I>.<p>
     * <B>File Input Statements:</B><BR>
     * If a line starts with "<<" the line is interpreted as a <I>file import
     * statement</I>.  A <I>file import statement</I> is of the form <<<filename
     * where the filename is the name of a file in the same directory as the
     * file being read.  For example, if properties file whose path is
     * "/home/username/foo.properties" contains the line "<<common.properties"
     * the properties in the file "/home/username/common.properties" are read
     * into the Properties object. File input statements may be recursive.
     * @return
     * java.util.Properties read from the file
     */
  	public static java.util.Properties loadPropertiesFile(String filePath)
   		throws java.io.IOException
   	{
    	return loadPropertiesFile(new java.io.File(filePath));
    }

    /*****
     * Loads the contents of a properties file which may contain
     * import statements for other properties files and returns a new
     * java.util.Properties object.  The properties file must be of the form
     * created by java.util.Properties.save(). The file is read using the same
     * rules as java.util.Properties.load() method with the addition of
     * <I>File Input Statements</I>.<p>
     * <B>File Input Statements:</B><BR>
     * If a line starts with "<<" the line is interpreted as a <I>file import
     * statement</I>.  A <I>file import statement</I> is of the form <<<filename
     * where the filename is the name of a file in the same directory as the
     * file being read.  For example, if properties file whose path is
     * "/home/username/foo.properties" contains the line "<<common.properties"
     * the properties in the file "/home/username/common.properties" are read
     * into the Properties object. File input statements may be recursive.
     * @return
     * java.util.Properties read from the file
     */
  	public static java.util.Properties loadPropertiesFile(java.io.File file)
   		throws java.io.IOException
   	{
    	return loadPropertiesFile(file, null);
    }

    /*****
     * Loads the contents of a properties file which may contain
     * import statements for other properties files and returns a new
     * java.util.Properties object.  The properties file must be of the form
     * created by java.util.Properties.save(). The file is read using the same
     * rules as java.util.Properties.load() method with the addition of
     * <I>File Input Statements</I>.<p>
     * <B>File Input Statements:</B><BR>
     * If a line starts with "<<" the line is interpreted as a <I>file import
     * statement</I>.  A <I>file import statement</I> is of the form <<<filename
     * where the filename is the name of a file in the same directory as the
     * file being read.  For example, if properties file whose path is
     * "/home/username/foo.properties" contains the line "<<common.properties"
     * the properties in the file "/home/username/common.properties" are read
     * into the Properties object. File input statements may be recursive.
     * @return
     * java.util.Properties read from the file
     */
  	public static java.util.Properties loadPropertiesFile(String filePath,
   		java.util.Properties defaults) throws java.io.IOException
    {
    	return loadPropertiesFile(new java.io.File(filePath), defaults);
    }


    /*****
     * Loads the contents of a properties file which may contain
     * import statements for other properties files and returns a new
     * java.util.Properties object.  The properties file must be of the form
     * created by java.util.Properties.save(). The file is read using the same
     * rules as java.util.Properties.load() method with the addition of
     * <I>File Input Statements</I>.<p>
     * <B>File Input Statements:</B><BR>
     * If a line starts with "<<" the line is interpreted as a <I>file import
     * statement</I>.  A <I>file import statement</I> is of the form <<<filename
     * where the filename is the name of a file in the same directory as the
     * file being read.  For example, if properties file whose path is
     * "/home/username/foo.properties" contains the line "<<common.properties"
     * the properties in the file "/home/username/common.properties" are read
     * into the Properties object. File input statements may be recursive.
     * @return
     * java.util.Properties read from the file
     */
  	public static java.util.Properties loadPropertiesFile(java.io.File file,
   		java.util.Properties defaults) throws java.io.IOException
   	{
    	StringBuffer b = loadProperties(file, null);
     	java.util.Properties p;
      	if (defaults == null) p = new java.util.Properties();
       	else p = new java.util.Properties(defaults);
        p.load(new java.io.ByteArrayInputStream(b.toString().getBytes()));
        return p;
    }

    /*****
     * recursive function to read properties files with import statements into
     * one long StringBuffer.  If buffer is null, a new StringBuffer is created,
     * otherwise, buffer is appended and returned.
     * @return
     * the same StringBuffer, buffer, with file's properties appended to it.
     */
    private static StringBuffer loadProperties(java.io.File file,
    	StringBuffer buffer) throws java.io.IOException
    {
    	if (buffer == null) buffer = new StringBuffer();
    	java.io.BufferedReader in = new java.io.BufferedReader(
     		new java.io.FileReader(file));
    	String ln = in.readLine();
     	while (ln != null)
      	{
       		if (ln.startsWith("<<"))
         	{
          		ln = ln.substring(2);
            	java.io.File parent = file.getParentFile();
             	java.io.File pf = new java.io.File(parent, ln);
          		buffer = loadProperties(pf, buffer);
            }
           	else buffer.append("\n" + ln);
         	ln = in.readLine();
       	}
        in.close();
        return buffer;
    }


    /***
     * Returns a FilenameFilter which will filter based on the extension
     * of the file.
     */
    public java.io.FilenameFilter createFilenameExtentionFilter(String ext)
    {
    	return new FileExtNameFilter(ext);
    }


    /***
     * Returns a FileFilter which will filter based on whether the file
     * is a directory or not
     */
    public java.io.FileFilter createFileDirFilter(boolean isDirectory)
    {
    	return new FileDirFilter(isDirectory);
    }

    /***
     * class to implement a filename filter based on an extension or
     * other suffix in the filename.
     */
    private class FileExtNameFilter implements java.io.FilenameFilter
    {
    	String ext;
    	FileExtNameFilter(String ext)
     	{
      		this.ext = ext;
      	}

        public boolean accept(java.io.File f, String name)
        {
        	return name.endsWith(ext);
        }
    }

    /***
     * class to implement a file filter based on whether file is directory
     * or not
     */
    private class FileDirFilter implements java.io.FileFilter
    {
    	private boolean isDir;
    	FileDirFilter(boolean isDirectory)
     	{
      		this.isDir = isDirectory;
      	}

        public boolean accept(java.io.File f)
        {
        	return f.isDirectory() ^ (!isDir);
        }
    }



}
