package webcore.webserver;

import webcore.utilities.*;
import java.net.*;
import java.io.*;
import java.util.*;

/*****
 * Class to hold the HTTPRequest from the browser.<p>
 * An instance of this class is passed to WebRequestables to handle the request.
 * There are methods to get the form variables and output stream from the
 * request.
 */
public class HTTPRequest
{
    public static final String HD_METHOD_GET = "GET";
    public static final String HD_METHOD_POST = "POST";
    public static final String HD_COOKIE = "COOKIE: ";
    public static final String HD_CONTENT_LEN = "CONTENT-LENGTH: ";
    public static final String HD_CONTENT_TYPE = "CONTENT-TYPE: ";
    public static final String HD_HOST = "HOST: ";
    public static final String HD_USERAGENT = "USER-AGENT: ";
    public static final String HD_MULTIPART_CONTENT = "multipart/form-data";
    public static final String HD_MULTIPART_BOUNDARY = "BOUNDARY=";
    public static final long MAX_CONTENT_LENGTH = 1024 * 10;

    private boolean processed = false;

    private String docBase;
    private Socket requestSocket;
    private String requestStr;

    private String methodStr;
    private String requestedObject;
    private String httpVersion;

    private String hostStr;
    private java.util.Hashtable formvarTable;
    private java.util.Hashtable cookieTable;
    private String userAgent;

    private boolean multipartData;
    private String multipartDataBorder;
    private char postData[] = null;
    private long contentLen;
    private String contentType;

    private java.io.PrintWriter pw;

    /****
     * constructs a new HTTPRequest after the request HTTP header has
     * been read from the stream.<p>
     * Post data has not been read yet.
     */
    public HTTPRequest(Socket requestSocket, String requestStr, String docBase)
        throws IOException
    {
        this.requestSocket = requestSocket;
        this.docBase = docBase;
        if (!this.docBase.endsWith(File.separator))
            this.docBase = this.docBase + File.separator;
        this.requestStr = requestStr;
        parseRequestStr();
    }

    /****
     * parses the request string and extract properties
     * of the request string
     */
    private void parseRequestStr() throws IOException
    {
        BufferedReader br = new BufferedReader(new StringReader(requestStr));
        String ln = br.readLine();
        String ucln;
        String cookies = null;
        String formVariables = null;
        formVariables = parseFirstLine(ln);
        while (ln != null)
        {
            ucln = ln.toUpperCase();

            if (ucln.startsWith(HD_COOKIE))
            {
                cookies = ln.substring(HD_COOKIE.length()).trim();
            }
            else
            if (ucln.startsWith(HD_CONTENT_LEN))
            {
                contentLen = Long.parseLong(
                    ln.substring(HD_CONTENT_LEN.length()).trim());
            }
            else
            if (ucln.startsWith(HD_CONTENT_TYPE))
            {
                contentType = ln.substring(HD_CONTENT_TYPE.length()).trim();
                if (contentType.startsWith(HD_MULTIPART_CONTENT))
                {
                    multipartDataBorder =
                        contentType.substring(HD_MULTIPART_CONTENT.length()
                            + 1); // +1 for semicolon.
                    if (multipartDataBorder.toUpperCase().startsWith(
                        HD_MULTIPART_BOUNDARY))
                        multipartDataBorder = multipartDataBorder.substring(
                            HD_MULTIPART_BOUNDARY.length()).trim();
                    contentType = HD_MULTIPART_CONTENT;
                }
            }
            else
            if (ucln.startsWith(HD_USERAGENT))
            {
                userAgent = ln.substring(HD_USERAGENT.length());
            }
            else
            if (ucln.startsWith(HD_HOST))
            {
                hostStr = ln.substring(HD_HOST.length());
            }
            ln = br.readLine();
        }
        if (methodStr.equals(HD_METHOD_POST) && !multipartData
            && contentLen < MAX_CONTENT_LENGTH)
        {
            byte postData[] = new byte[(int)contentLen];
            readFully(getRequestSocket().getInputStream(),
                postData);
            formVariables = new String(postData);
        }
        if (cookies != null) parseCookies(cookies);
        if (formVariables != null) parseFormVariables(formVariables);
    }


    /****
     * reads a full byte array from a stream.
     */
    private void readFully(InputStream in, byte b[]) throws IOException
    {
        int n = 0;
        int len = b.length;
        while (n < len)
        {
            int count = in.read(b, n, len - n);
            if (count < 0) throw new EOFException();
            n += count;
        }
    }

    /****
     * parses the first line of the request
     * @return
     * the form variables line if the method is GET and there
     * are form variables, otherwise null
     */
    private String parseFirstLine(String fl) throws IOException
    {
        StringTokenizer st = new StringTokenizer(fl);
        methodStr = st.nextToken();
        requestedObject = st.nextToken();
        String formVariables = null;
        if (methodStr.toUpperCase().equals(HD_METHOD_GET))
        {
            int i = requestedObject.indexOf('?');
            if (i >= 0)
            {
                formVariables = requestedObject.substring(i+1);
                requestedObject = requestedObject.substring(0, i);
            }
        }
        httpVersion = st.nextToken();
        return formVariables;
    }

    /***
     * parses the form variable line from the HTTP request
     */
    private void parseFormVariables(String formVariables)
    {
        formvarTable = new Hashtable();
        StringTokenizer st = new StringTokenizer(formVariables, "&");
        String fv;
        while (st.hasMoreElements())
        {
            fv = st.nextToken();
            StringTokenizer st2 = new StringTokenizer(fv, "=");
            try {formvarTable.put(st2.nextElement(), st2.nextElement());}
                catch(java.util.NoSuchElementException e) {}
        }
    }

    /****
     * parses the cookies line from the HTTP request
     */
    private void parseCookies(String cookies)
    {
        cookieTable = new Hashtable();
        StringTokenizer st = new StringTokenizer(cookies, "; ");
        String ck;
        String ckk, ckv;
        while (st.hasMoreElements())
        {
            ck = st.nextToken();
            StringTokenizer st2 = new StringTokenizer(ck, "=");
            ckk = st2.nextToken();
            ckv = st2.nextToken();
            cookieTable.put(ckk, ckv);
        }
     }


    public Socket getRequestSocket()
    {
        return requestSocket;
    }

    /***
     * returns a PrintWriter to write to the output stream for this request.
     */
    public java.io.PrintWriter getPrintWriter() throws java.io.IOException
    {
        if (pw == null)
            pw = new java.io.PrintWriter(requestSocket.getOutputStream());
        return pw;
    }

    /****
     * returns the OutputStream to write to for this request
     */
    public java.io.OutputStream getOutputStream() throws java.io.IOException
    {
        return requestSocket.getOutputStream();
    }


    /****
     * returns the InputStream to write to for this request
     */
    public java.io.InputStream getInputStream() throws java.io.IOException
    {
        return requestSocket.getInputStream();
    }

    /****
     * flushes output streams and closes the socket
     */
    public void close() throws java.io.IOException
    {
        getPrintWriter().flush();
        getPrintWriter().close();
        getRequestSocket().close();
    }

    private String switchSlashes(String path)
    {
        if (File.separatorChar == '/') return path;
        StringBuffer str = new StringBuffer(path);
        for (int i = 0; i < str.length(); i++)
        {
            if (str.charAt(i) == '/') str.setCharAt(i, File.separatorChar);
        }
        return str.toString();
    }



    /****
     * returns the document base, which is passed to the request
     * from the webserver
     */
    public String getDocBase()
    {
        return docBase;
    }

    /***
     * returns the URI
     */
    public String getRequestedObject()
    {
        return requestedObject;
    }

    /***
     * returns the URI with the slashes converted to the local filesystem
     */
    public String getRequestedFile()
    {
        return switchSlashes(getRequestedObject());
    }

    /***
     * returns full path to the requested file based on the document base
     */
    public String getRequestedPath()
    {
        return getDocBase() + getRequestedFile();
    }

    /****
     * returns the host passed in the HTTP header
     */
    public String getHost()
    {
        return hostStr;
    }

    /****
     * returns the method in the HTTP header (GET or POST are supported)
     */
    public String getMethod()
    {
        return methodStr;
    }




    /***
     * fixes common escape sequenses in variables
     */
    private String fixVar(String str)
    {
        String result = webcore.utilities.StringUtils.findReplace(str, "+", " ");
        int i = result.indexOf('%');
        while(i>=0 && result.length() > i + 2)
        {
            String chrstr = result.substring(i+1,i+3);
            char c;
            if (chrstr.charAt(0)=='%') c = (char)5;
            else c = (char)Integer.parseInt(chrstr, 16);
            if (c == '%') c = (char)5;
            result = webcore.utilities.StringUtils.findReplace(result, "%" + chrstr,
                ""+c);
            i = result.indexOf('%');
        }
        result = webcore.utilities.StringUtils.findReplace(result, ""+(char)5,"%");
        return result;
    }



    /****
     * returns all the form variables<p>
     * If the content type is multipart/form-data, the variables must be
     * read manually.
     */
    public String[] getFormVariableNames()
    {
        if (formvarTable == null) return new String []{};
        return
            new webcore.utilities.StringVector(
                formvarTable.keySet()).toStringArray();
    }

    /***
     * returns the string value of a form variable, if the variable is not
     * found, null is returned<p>
     * If the content type is multipart/form-data, the variables must be
     * read manually.
     */
    public String getFormVar(String formVariable)
    {
        if (formvarTable== null) return null;
        Object obj = formvarTable.get(formVariable);
        if (obj != null) return fixVar(obj.toString());
        else return null;
    }

    /*****
     * returns the cookie table.  If it was not created, it is created
     */
    private java.util.Hashtable getCookieTable()
    {
        return cookieTable;
    }

    /****
     * Returns an array containing all the available cookie names
     * provided by the request
     */
    public String[] getCookieNames()
    {
        java.util.Hashtable p = getCookieTable();
        if (p == null) return new String []{};
        StringVector v = new StringVector(p.keySet());
        return v.toStringArray();
    }

    /****
     * returns the value of a cookie.  If the cookie is not found,
     * null is returned.
     */
    public String getCookie(String cookieName)
    {
        if (getCookieTable() == null) return null;
        Object obj = getCookieTable().get(cookieName);
        return obj==null?null:fixVar(obj.toString());
    }

    /***
     * returns the request object string
     */
    public String toString()
    {
        return "HTTPRequest: " + requestedObject;
    }

    /***
     * determine if the request has been process or not
     */
    public boolean isProcessed()
    {
        return processed;
    }

    /****
     * if an event listener to the WebServer processes the request, it should
     * set the httpRequests processed property to TRUE using this method
     */
    public void setProcessed(boolean value)
    {
        processed = value;
    }

}
