package webcore.webserver;

import java.net.*;
import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.Runnable;
import java.util.*;

/****
 * WebServer listens on a port for HTTP requests and processes them.
 * <p>
 * The WebServer uses an instance of WebServerSocketFactoryInterface to
 * create listening Sockets.  The default factory creates instances of
 * java.net.ServerSocket.<p>
 * The WebServer uses an instance of WebRequestableFactoryInterface to
 * to process HTTP requests.
 */
public class WebServer implements Runnable
{
    private WebServerThreadPool threadPool = new WebServerThreadPool();
    private ServerSocket serverSock;
    private Thread runThread;
    private String documentBase = new String();
    private boolean terminated = true;
    private Vector listeners = new Vector();
    private int port = 80;
    private boolean ignoreClosedSocketException = false;
    private WebRequestableFactoryInterface webRequestableFactory =
        new DefaultWebRequestableFactory();
    private WebServerSocketFactoryInterface socketFactory =
        new DefaultWebServerSocketFactory();
    private static int MAX_THREADS = 25;
    private static int NORMAL_THREADS = 5;


    /****
     * Starts the webserver with the listening Thread being set as not daemon.
     * The webserver will continue to listen until terminate() is called.
     */
    public void start() throws IOException
    {
        start(false);
    }

    /***
     * Start the webserver. The webserver will continue to listen until
     * terminate() is called.
     * @param daemon	determine if the listening thread should be set as
     * a daemon thread or not
     */
    public void start(boolean daemon) throws IOException
    {
        if (!terminated) return;
        openServerSocket();
        MimeType.load();
        runThread = new Thread(this);
        runThread.setDaemon(daemon);
        terminated = false;
        runThread.start();
    }

    /****
     * returns the port being used
     */
    public int getPort()
    {
        return port;
    }

    /****
     * Sets the port the webserver will listen on.  The default HTTP port is
     * 80.  The port must be set before the WebServer is started or it will have
     * no affect.
     */
    public void setPort(int p)
    {
        port = p;
    }


    /****
     * determines if the webserver has been terminated
     */
    synchronized public boolean isTerminated()
    {
        return terminated;
    }

    /****
     * terminates the webserver, the compliment to start()
     */
    synchronized public void terminate() throws java.io.IOException
    {
        this.terminated = true;
        ignoreClosedSocketException = true;
        serverSock.close();
    }

    /****
     * returns the current web requestable factory.  This
     * factory is used to create WebRequestables to handle the HTTP requests.
     */
    public WebRequestableFactoryInterface getWebRequestableFactory()
    {
        return webRequestableFactory;
    }

    /****
     * sets the web requestable factory. This
     * factory is used to create WebRequestables to handle the HTTP requests.
     */
    public void setWebRequestableFactory(WebRequestableFactoryInterface factory)
    {
        webRequestableFactory = factory;
    }

    /****
     * returns the current socket factory. The factory is used to create
     * listening sockets for the server.
     */
    public WebServerSocketFactoryInterface getSocketFactory()
    {
        return socketFactory;
    }

    /***
     * The factory is used to create
     * listening sockets for the server. This factory must be set before the
     * start() method is called or it will have no affect.
     */
    public void setSocketFactory(WebServerSocketFactoryInterface factory)
    {
        socketFactory = factory;
    }

    /****
     * Sets the document base.  This is the path to the documents this
     * WebServer is serving.  This property is used to populate the
     * HTTPRequest's property. A WebRequestable should use the property of
     * HTTPRequest to determine where documents are located.
     * (HTTPFileRetriever used the property to locate the files it returns.)
     */
    public void setDocumentBase(String documentBase)
    {
        this.documentBase = documentBase;
    }

    /****
     * Gets the document base.  This is the path to the documents this
     * WebServer is serving.  This property is used to populate the
     * HTTPRequest's property. A WebRequestable should use the property of
     * HTTPRequest to determine where documents are located.
     * (HTTPFileRetriever used the property to locate the files it returns.)
     */
    public String getDocumentBase()
    {
        return documentBase;
    }

    /****
     * called by runThread to start the listening process.  This will
     * continuously accept connections and pass them to WebRequestables until
     * the WebServer has been terminated.
     */
    public void run()
    {
        Socket requestSock;
        while (!isTerminated())
        {
            requestSock = acceptSocket(serverSock);
            if (requestSock == null) continue;
            try
            {
                threadPool.processSocket(requestSock);
            }
            catch (Exception excpt)
            {
                fireMinorErrorOccured(excpt);
                try {requestSock.close();}
                catch (Exception e){fireMinorErrorOccured(e);}
            }
        }
        try
        {
            serverSock.close();
            fireStoppedListening("Stopped Listening.");
        }
        catch (Exception e)
        {
            fireMinorErrorOccured(e);
        }
    }




    /****
     * Method called by thread in Thread Pool to process an HTTP request.  This
     * method should not be called directly.  It reads the HTTP header,
     * constructs an HTTPRequest.  The method first passes the HTTPRequest to
     * all event listeners (webserver.WebServerListener).  If the HTTPRequest
     * is not processed, it is then processed by the WebRequestable returned
     * by the WebRequestableFactoryInterface of the WebServer instance.
     */
    private void processRequest(Socket requestSocket) throws IOException
    {
        InputStream in = requestSocket.getInputStream();
        ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        // read until there are two linebreaks in a row
        int lb = 0;
        int b;
        while (lb < 2)
        {
            b = in.read();
            if (b < 0) break;
            buffer.write(b);
            //if the char is \n, increase line break cnt.
            // if it is not \r or \n, reset line break cnt.
            if ((char)b=='\n') lb++;
            else
            if ((char)b=='\r') {}
            else
            lb=0;
        }
        String requestStr = new String(buffer.toByteArray());
        HTTPRequest httpRequest = new HTTPRequest(requestSocket,
            requestStr, documentBase);
        fireRequested(httpRequest);
        if (!httpRequest.isProcessed()) defaultProcessing(httpRequest);
    }


    /****
     * method called from processRequest() when no special processing is
     * required from a thread in the WebServerThreadPool
     */
    private void defaultProcessing(HTTPRequest httpRequest) throws IOException
    {
        try
        {
            WebRequestable wr =
                webRequestableFactory.createWebRequestable(httpRequest);
            if (wr != null) wr.processRequest(httpRequest);
            else throw new Exception("Null returned from factory");
        }
        catch (Exception e)
        {
            converge.kernel.Log.stack(e);
            //throw new IOException("Error using default processing: " +
            //   e.getMessage());
        }
    }

    /****
     * method when the WebServer is started to open the socket provided by
     * the socket factory
     */
    private void openServerSocket() throws IOException
    {
        serverSock = socketFactory.getServerSocket(port);
        fireStartedListening("Started listening on: " + port + " with " +
            socketFactory);
    }

    /*****
     * method called by run() to accept a socket
     */
    private Socket acceptSocket(ServerSocket sock)
    {
        try
        {
            return sock.accept();
        }
        catch (Exception e)
        {
            if (ignoreClosedSocketException)
            {
                ignoreClosedSocketException = false;
                return null;
            }
            fireMinorErrorOccured(e);
        }
        return null;
    }

    /****
     * adds a listener to this WebServer
     */
    public synchronized void addWebServerListener(WebServerListener l)
    {
        listeners.addElement(l);
    }

    /****
     * removes a listener from this WebServer
     */
    public synchronized void removeWebServerListener(WebServerListener l)
    {
        listeners.removeElement(l);
    }

    /****
     * internal method to fire event
     */
    private synchronized void fireStartedListening(Object arg)
    {
        for (int i=0; i<listeners.size(); i++)
            ((WebServerListener)listeners.elementAt(i)).startedListening(
                new WebServerEvent(this, arg));
    }

    /****
     * internal method to fire event
     */
    private synchronized void fireStoppedListening(Object arg)
    {
        for (int i=0; i<listeners.size(); i++)
            ((WebServerListener)listeners.elementAt(i)).stoppedListening(
                new WebServerEvent(this, arg));
    }

    /****
     * internal method to fire event
     */
    private synchronized void fireMinorErrorOccured(Object arg)
    {
        for (int i=0; i<listeners.size(); i++)
            ((WebServerListener)listeners.elementAt(i)).minorErrorOccured(
                new WebServerEvent(this, arg));
    }

    /****
     * internal method to fire event
     */
    private synchronized void fireRequested(Object arg)
    {
        for (int i=0; i<listeners.size(); i++)
            ((WebServerListener)listeners.elementAt(i)).requested(
                new WebServerEvent(this, arg));
    }


    /****
     * default factory for web requestables which are not in
     * the /java/ special area.
     */
    private static class DefaultWebRequestableFactory implements
        WebRequestableFactoryInterface
    {
        public WebRequestable createWebRequestable(HTTPRequest httpRequest)
        {
            if (httpRequest.getRequestedObject().startsWith("/java/"))
            {
                String cgiRequest = httpRequest.getRequestedObject().substring(
                    "/java/".length(),  httpRequest.getRequestedObject().length());
                int i = cgiRequest.indexOf("/");
                if (i < 0) i = cgiRequest.indexOf("?");
                if (i >= 0)	cgiRequest = cgiRequest.substring(0, i);
                try
                {
                    Class javaCGI = Class.forName(cgiRequest);
                    if (!WebRequestable.class.isAssignableFrom(javaCGI))
                        throw new Exception("Illegal server side java request");
                    WebRequestable requestable = (WebRequestable)javaCGI.newInstance();
                    return requestable;
                }
                catch (Exception e)
                {
                    return new HTTPFileRetriever();
                }
            }
            else return new HTTPFileRetriever();
        }
    }

    /****
     * default factory for server sockets
     */
    private static class DefaultWebServerSocketFactory implements
        WebServerSocketFactoryInterface
    {
        public java.net.ServerSocket getServerSocket(int requestedPort)
            throws java.io.IOException
        {
            return new java.net.ServerSocket(requestedPort);
        }
        public String toString()
        {
            return "Default Socket Factory";
        }
    }

    /***
     * simple class to hold a socket and a thread pool
     */
    private static class ThreadRequestStruct
    {
        Socket cli;
        WebServerThreadPool threadPool;
        ThreadRequestStruct(Socket cli, WebServerThreadPool threadPool)
        {
            this.cli = cli;
            this.threadPool = threadPool;
        }
    }


    /***
     * a thread for processing requests
     */
    private class ProcessRequestThread implements Runnable
    {

        private Object processLock = new Object();
        private java.util.LinkedList processList = new java.util.LinkedList();
        boolean running = true;

        public ProcessRequestThread(String name)
        {
            new Thread(this, name).start();
        }

        void processWebClient(Socket cli, WebServerThreadPool threadPool)
        {
            synchronized(processLock)
            {
                processList.add(new ThreadRequestStruct(cli, threadPool));
                processLock.notify();
            }
        }

        public void run()
        {
            while (running)
            {
                synchronized(processLock)
                {

                    ThreadRequestStruct rs = null;
                    try
                    {
                        try
                        {
                            rs = (ThreadRequestStruct) processList.removeFirst();
                            processRequest(rs.cli);
                        }
                        catch(java.util.NoSuchElementException e)
                        {
                            try{processLock.wait();}
                            catch (InterruptedException e3) {}
                        }
                        catch(Throwable e2)
                        {
                            fireMinorErrorOccured(e2);
                        }
                    }
                    finally
                    {
                        if (rs != null) rs.threadPool.replaceThread(this);
                    }
                }
            }
        }

    }

    /****
     * A pool for process request threads
     */
    private class WebServerThreadPool
    {
        java.util.Stack threadStack = new java.util.Stack();
        private int threadCnt=0;
        private static final int MAX_TRIES = 10;

        synchronized boolean processSocket(Socket cli)
        {
            ProcessRequestThread thread = null;
            int tryCnt = 0;
            do
            {
                try
                {
                    thread = (ProcessRequestThread)threadStack.pop();
                }
                catch (Exception e)
                {
                    if (threadCnt > 5)
                        try{wait(100);} catch (InterruptedException e2) {break;}
                }
            } while (thread == null && threadCnt > NORMAL_THREADS &&
                tryCnt++ < MAX_TRIES && threadCnt < MAX_THREADS);

            if (thread == null && threadCnt > MAX_THREADS)
            {
                try
                {
                    //TODO SEND TO BUSY MESSAGE
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
                return false;
            }

            if (thread == null)
            {
                thread = new ProcessRequestThread("ProcessRequestionThread: " +
                    threadCnt++);
            }
            thread.processWebClient(cli,this);
            return true;
        }

        synchronized void replaceThread(ProcessRequestThread t)
        {
            threadStack.push(t);
            notify();
        }
    }


}
